package jdk.internal.org.objectweb.asm.commons;

import java.util.Set;
import java.util.AbstractMap;
import jdk.internal.org.objectweb.asm.tree.InsnNode;
import jdk.internal.org.objectweb.asm.tree.LocalVariableNode;
import java.util.List;
import java.util.ArrayList;
import jdk.internal.org.objectweb.asm.tree.InsnList;
import java.util.LinkedList;
import jdk.internal.org.objectweb.asm.tree.AbstractInsnNode;
import jdk.internal.org.objectweb.asm.tree.LookupSwitchInsnNode;
import jdk.internal.org.objectweb.asm.tree.TableSwitchInsnNode;
import jdk.internal.org.objectweb.asm.tree.TryCatchBlockNode;
import java.util.Iterator;
import jdk.internal.org.objectweb.asm.tree.JumpInsnNode;
import jdk.internal.org.objectweb.asm.Label;
import java.util.HashMap;
import jdk.internal.org.objectweb.asm.MethodVisitor;
import java.util.BitSet;
import jdk.internal.org.objectweb.asm.tree.LabelNode;
import java.util.Map;
import jdk.internal.org.objectweb.asm.Opcodes;
import jdk.internal.org.objectweb.asm.tree.MethodNode;

public class JSRInlinerAdapter extends MethodNode implements Opcodes
{
    private static final boolean LOGGING = false;
    private final Map<LabelNode, BitSet> subroutineHeads;
    private final BitSet mainSubroutine;
    final BitSet dualCitizens;
    
    public JSRInlinerAdapter(MethodVisitor methodVisitor, int n, String s, String s2, String s3, String[] array) {
        this(327680, methodVisitor, n, s, s2, s3, array);
        if (this.getClass() != JSRInlinerAdapter.class) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected JSRInlinerAdapter(int api, MethodVisitor mv, int access, String name, String descriptor, String signature, String[] exceptions) {
        super(api, access, name, descriptor, signature, exceptions);
        this.subroutineHeads = new HashMap();
        this.mainSubroutine = new BitSet();
        this.dualCitizens = new BitSet();
        this.mv = mv;
    }
    
    @Override
    public void visitJumpInsn(int opcode, Label label) {
        super.visitJumpInsn(opcode, label);
        LabelNode label2 = ((JumpInsnNode)this.instructions.getLast()).label;
        if (opcode == 168 && !this.subroutineHeads.containsKey(label2)) {
            this.subroutineHeads.put(label2, new BitSet());
        }
    }
    
    @Override
    public void visitEnd() {
        if (!this.subroutineHeads.isEmpty()) {
            this.markSubroutines();
            this.emitCode();
        }
        if (this.mv != null) {
            this.accept(this.mv);
        }
    }
    
    private void markSubroutines() {
        BitSet set = new BitSet();
        this.markSubroutineWalk(this.mainSubroutine, 0, set);
        for (Map.Entry entry : this.subroutineHeads.entrySet()) {
            this.markSubroutineWalk((BitSet)entry.getValue(), this.instructions.indexOf((LabelNode)entry.getKey()), set);
        }
    }
    
    private void markSubroutineWalk(BitSet set, int n, BitSet set2) {
        this.markSubroutineWalkDFS(set, n, set2);
        int i = 1;
        while (i != 0) {
            i = 0;
            for (TryCatchBlockNode tryCatchBlockNode : this.tryCatchBlocks) {
                int index = this.instructions.indexOf(tryCatchBlockNode.handler);
                if (set.get(index)) {
                    continue;
                }
                int index2 = this.instructions.indexOf(tryCatchBlockNode.start);
                int index3 = this.instructions.indexOf(tryCatchBlockNode.end);
                int nextSetBit = set.nextSetBit(index2);
                if (nextSetBit != -1 && nextSetBit < index3) {
                    this.markSubroutineWalkDFS(set, index, set2);
                    i = 1;
                }
                else {
                    continue;
                }
            }
        }
    }
    
    private void markSubroutineWalkDFS(BitSet set, int index, BitSet set2) {
        while (true) {
            AbstractInsnNode value = this.instructions.get(index);
            if (set.get(index)) {
                return;
            }
            set.set(index);
            if (set2.get(index)) {
                this.dualCitizens.set(index);
            }
            set2.set(index);
            if (value.getType() == 7 && value.getOpcode() != 168) {
                this.markSubroutineWalkDFS(set, this.instructions.indexOf(((JumpInsnNode)value).label), set2);
            }
            if (value.getType() == 11) {
                TableSwitchInsnNode tableSwitchInsnNode = (TableSwitchInsnNode)value;
                this.markSubroutineWalkDFS(set, this.instructions.indexOf(tableSwitchInsnNode.dflt), set2);
                for (int i = tableSwitchInsnNode.labels.size() - 1; i >= 0; --i) {
                    this.markSubroutineWalkDFS(set, this.instructions.indexOf((LabelNode)tableSwitchInsnNode.labels.get(i)), set2);
                }
            }
            if (value.getType() == 12) {
                LookupSwitchInsnNode lookupSwitchInsnNode = (LookupSwitchInsnNode)value;
                this.markSubroutineWalkDFS(set, this.instructions.indexOf(lookupSwitchInsnNode.dflt), set2);
                for (int j = lookupSwitchInsnNode.labels.size() - 1; j >= 0; --j) {
                    this.markSubroutineWalkDFS(set, this.instructions.indexOf((LabelNode)lookupSwitchInsnNode.labels.get(j)), set2);
                }
            }
            switch (this.instructions.get(index).getOpcode()) {
                case 167:
                case 169:
                case 170:
                case 171:
                case 172:
                case 173:
                case 174:
                case 175:
                case 176:
                case 177:
                case 191: {
                    return;
                }
                default: {
                    if (++index >= this.instructions.size()) {
                        return;
                    }
                    continue;
                }
            }
        }
    }
    
    private void emitCode() {
        LinkedList<JSRInlinerAdapter.Instantiation> list = new LinkedList();
        list.add(new JSRInlinerAdapter.Instantiation(null, this.mainSubroutine));
        InsnList instructions = new InsnList();
        ArrayList<TryCatchBlockNode> tryCatchBlocks = new ArrayList();
        ArrayList<LocalVariableNode> localVariables = new ArrayList();
        while (!list.isEmpty()) {
            this.emitSubroutine((JSRInlinerAdapter.Instantiation)list.removeFirst(), list, instructions, tryCatchBlocks, localVariables);
        }
        this.instructions = instructions;
        this.tryCatchBlocks = tryCatchBlocks;
        this.localVariables = localVariables;
    }
    
    private void emitSubroutine(JSRInlinerAdapter.Instantiation previous, List<JSRInlinerAdapter.Instantiation> list, InsnList list2, List<TryCatchBlockNode> list3, List<LocalVariableNode> list4) {
        LabelNode labelNode = null;
        for (int i = 0; i < this.instructions.size(); ++i) {
            AbstractInsnNode value = this.instructions.get(i);
            JSRInlinerAdapter.Instantiation owner = previous.findOwner(i);
            if (value.getType() == 8) {
                LabelNode rangeLabel = previous.rangeLabel((LabelNode)value);
                if (rangeLabel != labelNode) {
                    list2.add(rangeLabel);
                    labelNode = rangeLabel;
                }
            }
            else if (owner == previous) {
                if (value.getOpcode() == 169) {
                    LabelNode returnLabel = null;
                    for (; previous != null; previous = previous.previous) {
                        if (previous.subroutine.get(i)) {
                            returnLabel = previous.returnLabel;
                        }
                    }
                    if (returnLabel == null) {
                        throw new RuntimeException("Instruction #" + i + " is a RET not owned by any subroutine");
                    }
                    list2.add(new JumpInsnNode(167, returnLabel));
                }
                else if (value.getOpcode() == 168) {
                    LabelNode label = ((JumpInsnNode)value).label;
                    JSRInlinerAdapter.Instantiation o = new JSRInlinerAdapter.Instantiation(previous, (BitSet)this.subroutineHeads.get(label));
                    LabelNode gotoLabel = o.gotoLabel(label);
                    list2.add(new InsnNode(1));
                    list2.add(new JumpInsnNode(167, gotoLabel));
                    list2.add(o.returnLabel);
                    list.add(o);
                }
                else {
                    list2.add(value.clone(previous));
                }
            }
        }
        for (TryCatchBlockNode tryCatchBlockNode : this.tryCatchBlocks) {
            LabelNode rangeLabel2 = previous.rangeLabel(tryCatchBlockNode.start);
            LabelNode rangeLabel3 = previous.rangeLabel(tryCatchBlockNode.end);
            if (rangeLabel2 == rangeLabel3) {
                continue;
            }
            LabelNode gotoLabel2 = previous.gotoLabel(tryCatchBlockNode.handler);
            if (rangeLabel2 == null || rangeLabel3 == null || gotoLabel2 == null) {
                throw new RuntimeException("Internal error!");
            }
            list3.add(new TryCatchBlockNode(rangeLabel2, rangeLabel3, gotoLabel2, tryCatchBlockNode.type));
        }
        for (LocalVariableNode localVariableNode : this.localVariables) {
            LabelNode rangeLabel4 = previous.rangeLabel(localVariableNode.start);
            LabelNode rangeLabel5 = previous.rangeLabel(localVariableNode.end);
            if (rangeLabel4 == rangeLabel5) {
                continue;
            }
            list4.add(new LocalVariableNode(localVariableNode.name, localVariableNode.desc, localVariableNode.signature, rangeLabel4, rangeLabel5, localVariableNode.index));
        }
    }
    
    private static void log(String x) {
        System.err.println(x);
    }
    
    private class Instantiation extends AbstractMap<LabelNode, LabelNode>
    {
        final JSRInlinerAdapter.Instantiation previous;
        public final BitSet subroutine;
        public final Map<LabelNode, LabelNode> rangeTable;
        public final LabelNode returnLabel;
        
        Instantiation(JSRInlinerAdapter.Instantiation previous, BitSet set) {
            this.rangeTable = new HashMap();
            this.previous = previous;
            this.subroutine = set;
            while (previous != null) {
                if (previous.subroutine == set) {
                    throw new RuntimeException("Recursive invocation of " + set);
                }
                previous = previous.previous;
            }
            if (previous != null) {
                this.returnLabel = new LabelNode();
            }
            else {
                this.returnLabel = null;
            }
            Object o = null;
            for (int i = 0; i < JSRInlinerAdapter.this.instructions.size(); ++i) {
                AbstractInsnNode value = JSRInlinerAdapter.this.instructions.get(i);
                if (value.getType() == 8) {
                    LabelNode labelNode = (LabelNode)value;
                    if (o == null) {
                        o = new LabelNode();
                    }
                    this.rangeTable.put(labelNode, o);
                }
                else if (this.findOwner(i) == this) {
                    o = null;
                }
            }
        }
        
        public JSRInlinerAdapter.Instantiation findOwner(int bitIndex) {
            if (!this.subroutine.get(bitIndex)) {
                return null;
            }
            if (!JSRInlinerAdapter.this.dualCitizens.get(bitIndex)) {
                return this;
            }
            for (JSRInlinerAdapter.Instantiation instantiation = this.previous; instantiation != null; instantiation = instantiation.previous) {
                if (instantiation.subroutine.get(bitIndex)) {
                    this = instantiation;
                }
            }
            return this;
        }
        
        public LabelNode gotoLabel(LabelNode insnNode) {
            return (LabelNode)this.findOwner(JSRInlinerAdapter.this.instructions.indexOf(insnNode)).rangeTable.get(insnNode);
        }
        
        public LabelNode rangeLabel(LabelNode labelNode) {
            return (LabelNode)this.rangeTable.get(labelNode);
        }
        
        @Override
        public Set<Map.Entry<LabelNode, LabelNode>> entrySet() {
            return null;
        }
        
        @Override
        public LabelNode get(Object o) {
            return this.gotoLabel((LabelNode)o);
        }
    }
}
